import logging
from typing import (
    Any,
    Dict,
    Optional,
    Tuple,
    Type,
    TypeAlias,
    Union,
)

from typing_extensions import (
    Self,
    assert_never,
)

from datahub.emitter.enum_helpers import get_enum_options
from datahub.errors import SdkUsageError
from datahub.metadata import schema_classes as models
from datahub.metadata.urns import (
    AssertionUrn,
    DatasetUrn,
    MonitorUrn,
    Urn,
)
from datahub.sdk.entity import Entity

logger = logging.getLogger(__name__)


MonitorIdentityInputType: TypeAlias = Union[
    Tuple[DatasetUrn, str],  # (dataset urn, monitor id)
    models.MonitorKeyClass,
    MonitorUrn,
    Tuple[
        DatasetUrn, AssertionUrn
    ],  # (dataset urn, assertion urn - monitor id is got from assertion id)
]

MonitorInfoInputType: TypeAlias = Union[
    Tuple[  # (monitor type [ASSERTION|FRESHNESS], monitor status mode [ACTIVE|INACTIVE|PASSIVE])
        str, str
    ],
    Tuple[  # (monitor type enum, monitor status mode enum)
        models.MonitorTypeClass, models.MonitorModeClass
    ],
    models.MonitorInfoClass,
]


class Monitor(Entity):
    """
    Monitor entity class.
    """

    __slots__ = ()

    @classmethod
    def get_urn_type(cls) -> Type[MonitorUrn]:
        """Get the URN type for monitors.

        Returns:
            The MonitorUrn class.
        """
        return MonitorUrn

    def __init__(
        self,
        id: MonitorIdentityInputType,
        # MonitorInfo
        info: MonitorInfoInputType,
        external_url: Optional[str] = None,
        custom_properties: Optional[Dict[str, str]] = None,
    ) -> None:
        super().__init__(urn=Monitor._ensure_id(id=id))

        self._set_info(info)

        if external_url is not None:
            # this may overwrite the value set from info#externalUrl value
            self.set_external_url(external_url)
        if custom_properties is not None:
            # this may overwrite the value set from info#customPropperties value
            self.set_custom_properties(custom_properties)

    @classmethod
    def _new_from_graph(cls, urn: Urn, current_aspects: models.AspectBag) -> Self:
        assert isinstance(urn, MonitorUrn)
        assert "monitorInfo" in current_aspects, "MonitorInfo is required"

        entity = cls(id=urn, info=current_aspects["monitorInfo"])
        return entity._init_from_graph(current_aspects)

    @property
    def urn(self) -> MonitorUrn:
        assert isinstance(self._urn, MonitorUrn)
        return self._urn

    @classmethod
    def _ensure_id(
        cls,
        id: MonitorIdentityInputType,
    ) -> MonitorUrn:
        if isinstance(id, tuple):
            if (
                len(id) != 2
                or not isinstance(id[0], DatasetUrn)
                or not isinstance(id[1], (str, AssertionUrn))
            ):
                raise SdkUsageError(
                    f"Invalid monitor identity input tuple: {id}. Expected a tuple of (dataset/entity urn, monitor id)."
                )
            if isinstance(id[1], str):
                # If the second element is a string, we treat it as the monitor id.
                # This allows for flexibility in how monitor ids are specified.
                return MonitorUrn(entity=id[0], id=id[1])
            elif isinstance(id[1], AssertionUrn):
                # If the second element is an AssertionUrn, we use assertion urn as the monitor id so they both are linked.
                return MonitorUrn(entity=id[0], id=id[1].urn())
            else:
                assert_never(id[1])
        elif isinstance(id, models.MonitorKeyClass):
            # This validation may look redundant but it is not.
            # While MonitorKey PDL model ensures entity is a DatasetUrn,
            # generated python code for entity's type in MonitorKeyClass is not so strict and allows for any string.
            entity: DatasetUrn
            if isinstance(id.entity, str):
                try:
                    entity = DatasetUrn.from_string(id.entity)
                except Exception as e:
                    raise SdkUsageError(
                        f"Invalid monitor identity input key: {id}. Invalid dataset urn for entity."
                    ) from e
            else:
                if not isinstance(id.entity, DatasetUrn):
                    raise SdkUsageError(
                        f"Invalid monitor identity input key: {id}. Expected a DatasetUrn for entity."
                    )
                entity = id.entity
            return MonitorUrn(entity=entity, id=id.id)
        elif isinstance(id, MonitorUrn):
            return id
        else:
            assert_never(id)

    def _ensure_info(self) -> models.MonitorInfoClass:
        return self._setdefault_aspect(
            models.MonitorInfoClass(
                # TBC: just because we need to set them!
                type=models.MonitorTypeClass.ASSERTION,
                status=models.MonitorStatusClass(mode=models.MonitorModeClass.ACTIVE),
            )
        )

    @property
    def info(self) -> models.MonitorInfoClass:
        """Get the monitor info.

        Returns:
            The monitor info.
        """
        return self._ensure_info()

    def _set_info(self, info: MonitorInfoInputType) -> None:
        """Set the monitor info.

        Args:
            info: The monitor info.
        """
        if isinstance(info, tuple):
            if len(info) != 2:
                raise SdkUsageError(
                    f"Invalid monitor info input tuple: {info}. Expected a tuple of (monitor type, monitor status)."
                )
            type, mode_status = info
            if isinstance(type, str) and type not in get_enum_options(
                models.MonitorTypeClass
            ):
                raise SdkUsageError(
                    f"Invalid monitor type: {type}. Expected {get_enum_options(models.MonitorTypeClass)}."
                )
            if isinstance(mode_status, str) and mode_status not in get_enum_options(
                models.MonitorModeClass
            ):
                raise SdkUsageError(
                    f"Invalid monitor status: {mode_status}. Expected {get_enum_options(models.MonitorModeClass)}."
                )
            self._ensure_info().type = type
            self._ensure_info().status = models.MonitorStatusClass(mode=mode_status)
        elif isinstance(info, models.MonitorInfoClass):
            self._set_aspect(info)
        else:
            assert_never(info)

    @property
    def external_url(self) -> Optional[str]:
        """Get the external URL of the monitor.

        Returns:
            The external URL if set, None otherwise.
        """
        return self._ensure_info().externalUrl

    def set_external_url(self, external_url: str) -> None:
        """Set the external URL of the monitor.

        Args:
            external_url: The external URL to set.
        """
        self._ensure_info().externalUrl = external_url

    @property
    def custom_properties(self) -> Dict[str, str]:
        """Get the custom properties of the monitor.

        Returns:
            Dictionary of custom properties.
        """
        return self._ensure_info().customProperties

    def set_custom_properties(self, custom_properties: Dict[str, str]) -> None:
        """Set the custom properties of the monitor.

        Args:
            custom_properties: Dictionary of custom properties to set.
        """
        self._ensure_info().customProperties = custom_properties

    @property
    def sensitivity(self) -> Optional[models.AssertionMonitorSensitivityClass]:
        """Get the sensitivity of the monitor.

        Returns:
            The sensitivity or None if not found.
        """
        return _get_nested_field_for_entity_with_default(
            self, "info.assertionMonitor.settings.adjustmentSettings.sensitivity"
        )

    @property
    def exclusion_windows(self) -> Optional[list[models.AssertionExclusionWindowClass]]:
        """Get the exclusion windows of the monitor.

        Returns:
            The exclusion windows or None if not found.
        """
        return _get_nested_field_for_entity_with_default(
            self, "info.assertionMonitor.settings.adjustmentSettings.exclusionWindows"
        )

    @property
    def training_data_lookback_days(self) -> Optional[int]:
        """Get the training data lookback days of the monitor.

        Returns:
            The training data lookback days or None if not found.
        """
        return _get_nested_field_for_entity_with_default(
            self,
            "info.assertionMonitor.settings.adjustmentSettings.trainingDataLookbackWindowDays",
        )


def _get_nested_field_for_entity_with_default(
    entity: Entity,
    field_path: str,
    default: Any = None,
) -> Any:
    """
    Get a nested field from an Entity object, and warn and return default if not found.

    Args:
        entity: The entity to get the nested field from.
        field_path: The path to the nested field.
        default: The default value to return if the field is not found.
    """
    fields = field_path.split(".")
    current = entity
    last_valid_path = entity.entity_type_name()

    for field in fields:
        try:
            current = getattr(current, field)
            last_valid_path = f"{last_valid_path}.{field}"
        except AttributeError:
            logger.warning(
                f"{entity.entity_type_name().capitalize()} {entity.urn} does not have an `{last_valid_path}` field, defaulting to {default}"
            )
            return default

    return current



from typing import Any, Dict, Optional, Union

from datahub.ingestion.graph.client import DataHubGraph


def _parse_response(response: Dict) -> Dict:
    value_str = response["value"].strip("{}")

    args_part = value_str.split(", result=")[0].replace("args=", "")
    result_part = value_str.split(", result=")[1]

    parsed = {"args": _parse_args(args_part), "result": _parse_result(result_part)}
    return parsed


def _parse_dict(content: str) -> Dict:
    content = content.strip("()").strip()
    items = [item.strip() for item in content.split(", ")]

    pairs = []
    for item in items:
        if "=" in item:
            k, v = item.split("=", 1)
            pairs.append((k.strip(), v.strip()))

    return {k: _convert_value(v) for k, v in pairs}


def _parse_args(args_str: str) -> Dict:
    content = args_str.replace("RestoreIndicesArgs(", "").replace(")", "").strip()
    return _parse_dict(content)


def _parse_result(result_str: str) -> Dict:
    content = result_str.replace("RestoreIndicesResult(", "").replace(")", "").strip()
    return _parse_dict(content)


def _convert_value(v: str) -> Any:
    if v == "null":
        return None
    elif v.isdigit():
        return int(v)
    elif v == "[]":
        return []
    elif v == "":
        return ""
    elif v == "false":
        return False
    elif v == "true":
        return True
    elif v.startswith("urn:"):
        return v
    return v


def restore_indices(
    graph: DataHubGraph,
    start: int,
    batch_size: int,
    limit: int,
    urn: Optional[str] = None,
    urn_like: Optional[str] = None,
    aspect: Optional[str] = None,
) -> Dict:
    if urn is None and urn_like is None:
        raise RuntimeError("Either urn or urn_like must be present")

    url = f"{graph.config.server}/operations?action=restoreIndices"
    payload_dict: Dict[str, Union[str, int]] = {
        "start": start,
        "batchSize": batch_size,
        "limit": limit,
    }
    if urn_like is not None:
        payload_dict["urnLike"] = urn_like
    if urn is not None:
        payload_dict["urn"] = urn
    if aspect is not None:
        payload_dict["aspect"] = aspect
    response = graph._post_generic(
        url=url,
        payload_dict=payload_dict,
    )
    return _parse_response(response)



import uuid
from datetime import datetime
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    TypeAlias,
    Union,
)

from typing_extensions import (
    Self,
    assert_never,
)

from acryl_datahub_cloud.sdk.errors import SDKNotYetSupportedError
from datahub.emitter.enum_helpers import get_enum_options
from datahub.emitter.mce_builder import make_ts_millis
from datahub.errors import SdkUsageError
from datahub.metadata import schema_classes as models
from datahub.metadata.urns import AssertionUrn, DatasetUrn
from datahub.sdk._shared import (
    HasPlatformInstance,
    HasTags,
    TagsInputType,
)
from datahub.sdk.entity import Entity
from datahub.utilities.urns._urn_base import Urn

AssertionInfoInputType: TypeAlias = Union[
    models.DatasetAssertionInfoClass,
    models.FreshnessAssertionInfoClass,
    models.VolumeAssertionInfoClass,
    models.SqlAssertionInfoClass,
    models.FieldAssertionInfoClass,
    # TODO: models.SchemaAssertionInfoClass,
    # TODO: models.CustomAssertionInfoClass,
]
AssertionSourceInputType: TypeAlias = Union[
    str,  # assertion source type
    models.AssertionSourceTypeClass,
    models.AssertionSourceClass,
]
LastUpdatedInputType: TypeAlias = Union[
    Tuple[int, str],  # (timestamp, actor)
    Tuple[datetime, str],  # (datetime, actor)
    models.AuditStampClass,
]
AssertionActionsInputType: TypeAlias = List[
    Union[
        str,  # assertion action type: RAISE_INCIDENT, RESOLVE_INCIDENT
        models.AssertionActionTypeClass,
        models.AssertionActionClass,
    ]
]


class Assertion(HasPlatformInstance, HasTags, Entity):
    """Represents an assertion in DataHub.

    TODO: This is a placeholder for better documentation.
    """

    __slots__ = ()

    @classmethod
    def get_urn_type(cls) -> Type[AssertionUrn]:
        """Get the URN type for assertions.

        Returns:
            The AssertionUrn class.
        """
        return AssertionUrn

    def __init__(
        self,
        *,
        # Identity; it is automatically generated if not provided
        id: Union[str, AssertionUrn, None] = None,
        # Assertion info
        info: AssertionInfoInputType,
        description: Optional[str] = None,
        external_url: Optional[str] = None,
        custom_properties: Optional[Dict[str, str]] = None,
        source: Optional[AssertionSourceInputType] = None,
        last_updated: Optional[LastUpdatedInputType] = None,
        # For assertions, platform+instance is not coupled to identity
        platform: Optional[str] = None,
        platform_instance: Optional[str] = None,
        # Standard aspects
        tags: Optional[TagsInputType] = None,
        # Assertion actions
        on_success: Optional[AssertionActionsInputType] = None,
        on_failure: Optional[AssertionActionsInputType] = None,
    ):
        super().__init__(urn=Assertion._ensure_id(id))

        self._set_info(info)
        if description is not None:
            self.set_description(description)
        if external_url is not None:
            self.set_external_url(external_url)
        if custom_properties is not None:
            self.set_custom_properties(custom_properties)
        if source is not None:
            self._set_source(source)
        if last_updated is not None:
            self._set_last_updated(last_updated)

        if platform is not None:
            self._set_platform_instance(platform, platform_instance)

        if tags is not None:
            self.set_tags(tags)

        if on_success is not None or on_failure is not None:
            self._set_actions(on_success or [], on_failure or [])

    @classmethod
    def _new_from_graph(cls, urn: Urn, current_aspects: models.AspectBag) -> Self:
        assert isinstance(urn, AssertionUrn)
        assert "assertionInfo" in current_aspects, "AssertionInfo is required"
        assertion_info = current_aspects["assertionInfo"]
        info = cls._switch_assertion_info(assertion_info)

        entity = cls(id=urn, info=info)
        return entity._init_from_graph(current_aspects)

    @property
    def urn(self) -> AssertionUrn:
        assert isinstance(self._urn, AssertionUrn)
        return self._urn

    @classmethod
    def _ensure_id(cls, id: Union[str, AssertionUrn, None]) -> AssertionUrn:
        if isinstance(id, str):
            return AssertionUrn.from_string(id)
        elif isinstance(id, AssertionUrn):
            return id
        elif id is None:
            return AssertionUrn.from_string(f"urn:li:assertion:{uuid.uuid4()}")
        else:
            assert_never(id)

    def _ensure_info(self) -> models.AssertionInfoClass:
        # Compared with datasets, we don't have editable non-editable duality here
        return self._setdefault_aspect(
            models.AssertionInfoClass(
                # TBC: just because we need to set one!
                type=models.AssertionTypeClass.DATASET
            )
        )

    def _ensure_actions(self) -> models.AssertionActionsClass:
        return self._setdefault_aspect(
            models.AssertionActionsClass(onSuccess=[], onFailure=[])
        )

    @property
    def description(self) -> Optional[str]:
        """Get the description of the assertion.

        Returns:
            The descripton if set, None otherwise.
        """
        return self._ensure_info().description

    def set_description(self, description: str) -> None:
        """Set the description of the assertion.

        Args:
            description: The description to set.
        """
        self._ensure_info().description = description

    def _set_info(self, assertion: AssertionInfoInputType) -> None:
        info = self._ensure_info()
        if isinstance(assertion, models.DatasetAssertionInfoClass):
            info.datasetAssertion = assertion
            info.type = models.AssertionTypeClass.DATASET
        elif isinstance(assertion, models.FreshnessAssertionInfoClass):
            info.freshnessAssertion = assertion
            info.type = models.AssertionTypeClass.FRESHNESS
        elif isinstance(assertion, models.VolumeAssertionInfoClass):
            info.volumeAssertion = assertion
            info.type = models.AssertionTypeClass.VOLUME
        elif isinstance(assertion, models.SqlAssertionInfoClass):
            info.sqlAssertion = assertion
            info.type = models.AssertionTypeClass.SQL
        elif isinstance(assertion, models.FieldAssertionInfoClass):
            info.fieldAssertion = assertion
            info.type = models.AssertionTypeClass.FIELD
        else:
            assert_never(assertion)

    @property
    def info(self) -> AssertionInfoInputType:
        """Get the assertion info.

        Returns:
            The assertion info.
        """
        info = self._ensure_info()
        return Assertion._switch_assertion_info(info)

    @property
    def external_url(self) -> Optional[str]:
        """Get the external URL of the assertion.

        Returns:
            The external URL if set, None otherwise.
        """
        return self._ensure_info().externalUrl

    def set_external_url(self, external_url: str) -> None:
        """Set the external URL of the assertion.

        Args:
            external_url: The external URL to set.
        """
        self._ensure_info().externalUrl = external_url

    @property
    def custom_properties(self) -> Dict[str, str]:
        """Get the custom properties of the assertion.

        Returns:
            Dictionary of custom properties.
        """
        return self._ensure_info().customProperties

    def set_custom_properties(self, custom_properties: Dict[str, str]) -> None:
        """Set the custom properties of the assertion.

        Args:
            custom_properties: Dictionary of custom properties to set.
        """
        self._ensure_info().customProperties = custom_properties

    @property
    def source(self) -> Optional[models.AssertionSourceClass]:
        """Get the source of the assertion.

        Returns:
            The source if set, None otherwise.
        """
        return self._ensure_info().source

    def _set_source(self, source: AssertionSourceInputType) -> None:
        """Set the source of the assertion.

        Args:
            source: The source to set.
        """
        info = self._ensure_info()
        if isinstance(source, str):
            # TBC: The model allows arbitrary strings, should we restrict this to AssertionSourceTypeClass enum values?
            info.source = models.AssertionSourceClass(
                type=source,
            )
        elif isinstance(source, models.AssertionSourceTypeClass):
            info.source = models.AssertionSourceClass(
                type=source,
            )
        elif isinstance(source, models.AssertionSourceClass):
            info.source = source
        else:
            assert_never(source)

    @property
    def last_updated(self) -> Optional[models.AuditStampClass]:
        """Get the last updated audit event of the assertion.

        Audit event includes the timestamp and the actor who made the change.

        Returns:
            The last updated audit event of the assertion, or None if not set.
        """
        return self._ensure_info().lastUpdated

    def _set_last_updated(self, last_updated: LastUpdatedInputType) -> None:
        """Set the last updated audit event, including the timestamp and the actor who made the change.

        Args:
            last_updated: A tuple containing the last updated timestamp and actor.
        """
        info = self._ensure_info()

        if isinstance(last_updated, tuple):
            if len(last_updated) != 2:
                raise SdkUsageError("Invalid length for last updated tuple")

            tm, actor = last_updated

            if not self._is_actor_urn(actor):
                raise SdkUsageError(
                    "Invalid actor for last updated tuple, expected 'urn:li:corpuser:*' or 'urn:li:corpGroup:*'"
                )

            if isinstance(tm, datetime):
                info.lastUpdated = models.AuditStampClass(make_ts_millis(tm), actor)
            elif isinstance(tm, int):
                info.lastUpdated = models.AuditStampClass(tm, actor)
            else:
                raise SdkUsageError(
                    "Invalid type for last updated tuple timestamp, expected int or datetime"
                )
        elif isinstance(last_updated, models.AuditStampClass):
            if not self._is_actor_urn(last_updated.actor):
                raise SdkUsageError(
                    "Invalid actor for last updated tuple, expected 'urn:li:corpuser:*' or 'urn:li:corpGroup:*'"
                )
            info.lastUpdated = last_updated
        else:
            assert_never(last_updated)

    # TODO: is this defined somewhere else?
    def _is_actor_urn(self, actor: str) -> bool:
        """Check if the actor is a valid URN.

        Args:
            actor: The actor to check.

        Returns:
            True if the actor is a valid URN, False otherwise.
        """
        return actor.startswith("urn:li:corpuser:") or actor.startswith(
            "urn:li:corpGroup:"
        )

    @property
    def on_success(self) -> List[models.AssertionActionClass]:
        """Get the actions to perform on success.

        Returns:
            The actions to perform on success, it may be empty if not set.
        """
        return self._ensure_actions().onSuccess

    @property
    def on_failure(self) -> List[models.AssertionActionClass]:
        """Get the actions to perform on failure.

        Returns:
            The actions to perform on failure, it may be empty if not set.
        """
        return self._ensure_actions().onFailure

    def _set_actions(
        self,
        on_success: AssertionActionsInputType,
        on_failure: AssertionActionsInputType,
    ) -> None:
        """Set the actions to perform on success or failure.

        Args:
            on_success: The actions to perform on success.
            on_failure: The actions to perform on failure.
        """
        for action in on_success + on_failure:
            if isinstance(action, str) and action not in [
                models.AssertionActionTypeClass.RAISE_INCIDENT,
                models.AssertionActionTypeClass.RESOLVE_INCIDENT,
            ]:
                raise SdkUsageError(
                    f"Invalid action type {str} for on_success or on_failure actions, expected valid str or AssertionActionTypeClass"
                )

        actions = self._ensure_actions()
        actions.onSuccess = [
            action
            if isinstance(action, models.AssertionActionClass)
            else models.AssertionActionClass(type=action)
            for action in on_success
        ]
        actions.onFailure = [
            action
            if isinstance(action, models.AssertionActionClass)
            else models.AssertionActionClass(type=action)
            for action in on_failure
        ]

    @property
    def dataset(self) -> DatasetUrn:
        """Get the dataset URN associated with the assertion.

        Returns:
            The dataset URN.
        """
        info = self._ensure_info()
        assertion_info = Assertion._switch_assertion_info(info)

        dataset_str = (
            assertion_info.dataset
            if isinstance(assertion_info, models.DatasetAssertionInfoClass)
            else assertion_info.entity
        )
        return DatasetUrn.from_string(dataset_str)

    @classmethod
    def _switch_assertion_info(
        cls,
        assertion_info: models.AssertionInfoClass,
    ) -> Union[
        models.DatasetAssertionInfoClass,
        models.FreshnessAssertionInfoClass,
        models.VolumeAssertionInfoClass,
        models.SqlAssertionInfoClass,
        models.FieldAssertionInfoClass,
    ]:
        if assertion_info.type not in get_enum_options(models.AssertionTypeClass):
            raise SDKNotYetSupportedError(f"Assertion type: {assertion_info.type}")

        if assertion_info.type == models.AssertionTypeClass.DATASET:
            assert assertion_info.datasetAssertion is not None
            return assertion_info.datasetAssertion
        elif assertion_info.type == models.AssertionTypeClass.FRESHNESS:
            assert assertion_info.freshnessAssertion is not None
            return assertion_info.freshnessAssertion
        elif assertion_info.type == models.AssertionTypeClass.VOLUME:
            assert assertion_info.volumeAssertion is not None
            return assertion_info.volumeAssertion
        elif assertion_info.type == models.AssertionTypeClass.SQL:
            assert assertion_info.sqlAssertion is not None
            return assertion_info.sqlAssertion
        elif assertion_info.type == models.AssertionTypeClass.FIELD:
            assert assertion_info.fieldAssertion is not None
            return assertion_info.fieldAssertion
        else:
            raise AssertionError("Unreachable code, all cases should be handled above")



# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import AssetCollectionModuleParamsClass
from .....schema_classes import DataHubPageModuleParamsClass
from .....schema_classes import DataHubPageModulePropertiesClass
from .....schema_classes import DataHubPageModuleTypeClass
from .....schema_classes import DataHubPageModuleVisibilityClass
from .....schema_classes import HierarchyModuleParamsClass
from .....schema_classes import LinkModuleParamsClass
from .....schema_classes import PageModuleScopeClass
from .....schema_classes import RichTextModuleParamsClass


AssetCollectionModuleParams = AssetCollectionModuleParamsClass
DataHubPageModuleParams = DataHubPageModuleParamsClass
DataHubPageModuleProperties = DataHubPageModulePropertiesClass
DataHubPageModuleType = DataHubPageModuleTypeClass
DataHubPageModuleVisibility = DataHubPageModuleVisibilityClass
HierarchyModuleParams = HierarchyModuleParamsClass
LinkModuleParams = LinkModuleParamsClass
PageModuleScope = PageModuleScopeClass
RichTextModuleParams = RichTextModuleParamsClass

# fmt: on



from datetime import datetime
from enum import Enum
from typing import Optional, Union

from pydantic import BaseModel

from acryl_datahub_cloud.sdk.assertion_input.assertion_input import (
    DEFAULT_EVERY_SIX_HOURS_SCHEDULE,
    RANGE_OPERATORS,
    SINGLE_VALUE_NUMERIC_OPERATORS,
    AssertionIncidentBehaviorInputTypes,
    FieldSpecType,
    _AssertionInput,
)
from acryl_datahub_cloud.sdk.entities.assertion import (
    AssertionInfoInputType,
    TagsInputType,
)
from acryl_datahub_cloud.sdk.errors import SDKUsageError
from datahub.metadata import schema_classes as models
from datahub.metadata.urns import AssertionUrn, CorpUserUrn, DatasetUrn
from datahub.sdk.entity_client import EntityClient


class SqlAssertionCondition(Enum):
    IS_EQUAL_TO = "IS_EQUAL_TO"  # models.SqlAssertionTypeClass.METRIC + models.AssertionStdOperatorClass.EQUAL_TO
    IS_NOT_EQUAL_TO = "IS_NOT_EQUAL_TO"  # models.SqlAssertionTypeClass.METRIC + models.AssertionStdOperatorClass.NOT_EQUAL_TO
    IS_GREATER_THAN = "IS_GREATER_THAN"  # models.SqlAssertionTypeClass.METRIC + models.AssertionStdOperatorClass.GREATER_THAN
    IS_LESS_THAN = "IS_LESS_THAN"  # models.SqlAssertionTypeClass.METRIC + models.AssertionStdOperatorClass.LESS_THAN
    IS_WITHIN_A_RANGE = "IS_WITHIN_A_RANGE"  # models.SqlAssertionTypeClass.METRIC + models.AssertionStdOperatorClass.BETWEEN
    GROWS_AT_MOST_ABSOLUTE = "GROWS_AT_MOST_ABSOLUTE"  # models.SqlAssertionTypeClass.METRIC_CHANGE + models.AssertionStdOperatorClass.LESS_THAN_OR_EQUAL_TO + models.AssertionValueChangeTypeClass.ABSOLUTE
    GROWS_AT_MOST_PERCENTAGE = "GROWS_AT_MOST_PERCENTAGE"  # models.SqlAssertionTypeClass.METRIC_CHANGE + models.AssertionStdOperatorClass.LESS_THAN_OR_EQUAL_TO + models.AssertionValueChangeTypeClass.PERCENTAGE
    GROWS_AT_LEAST_ABSOLUTE = "GROWS_AT_LEAST_ABSOLUTE"  # models.SqlAssertionTypeClass.METRIC_CHANGE + models.AssertionStdOperatorClass.GREATER_THAN_OR_EQUAL_TO + models.AssertionValueChangeTypeClass.ABSOLUTE
    GROWS_AT_LEAST_PERCENTAGE = "GROWS_AT_LEAST_PERCENTAGE"  # models.SqlAssertionTypeClass.METRIC_CHANGE + models.AssertionStdOperatorClass.GREATER_THAN_OR_EQUAL_TO + models.AssertionValueChangeTypeClass.PERCENTAGE
    GROWS_WITHIN_A_RANGE_ABSOLUTE = "GROWS_WITHIN_A_RANGE_ABSOLUTE"  # models.SqlAssertionTypeClass.METRIC_CHANGE + models.AssertionStdOperatorClass.BETWEEN + models.AssertionValueChangeTypeClass.ABSOLUTE
    GROWS_WITHIN_A_RANGE_PERCENTAGE = "GROWS_WITHIN_A_RANGE_PERCENTAGE"  # models.SqlAssertionTypeClass.METRIC_CHANGE + models.AssertionStdOperatorClass.BETWEEN + models.AssertionValueChangeTypeClass.PERCENTAGE


class SqlAssertionCriteria(BaseModel):
    condition: Union[SqlAssertionCondition, str]
    # Either a single value or a range must be provided
    parameters: Union[Union[float, int], tuple[Union[float, int], Union[float, int]]]

    @classmethod
    def get_type_from_condition(
        cls, condition: Union[SqlAssertionCondition, str]
    ) -> models.SqlAssertionTypeClass:
        condition_enum = (
            condition
            if isinstance(condition, SqlAssertionCondition)
            else SqlAssertionCondition(condition)
        )

        if condition_enum in [
            SqlAssertionCondition.IS_EQUAL_TO,
            SqlAssertionCondition.IS_NOT_EQUAL_TO,
            SqlAssertionCondition.IS_GREATER_THAN,
            SqlAssertionCondition.IS_LESS_THAN,
            SqlAssertionCondition.IS_WITHIN_A_RANGE,
        ]:
            return models.SqlAssertionTypeClass.METRIC  # type: ignore[return-value]
        elif condition_enum in [
            SqlAssertionCondition.GROWS_AT_MOST_ABSOLUTE,
            SqlAssertionCondition.GROWS_AT_MOST_PERCENTAGE,
            SqlAssertionCondition.GROWS_AT_LEAST_ABSOLUTE,
            SqlAssertionCondition.GROWS_AT_LEAST_PERCENTAGE,
            SqlAssertionCondition.GROWS_WITHIN_A_RANGE_ABSOLUTE,
            SqlAssertionCondition.GROWS_WITHIN_A_RANGE_PERCENTAGE,
        ]:
            return models.SqlAssertionTypeClass.METRIC_CHANGE  # type: ignore[return-value]
        else:
            raise SDKUsageError(f"Unknown condition: {condition_enum}")

    @classmethod
    def get_operator_from_condition(
        cls, condition: Union[SqlAssertionCondition, str]
    ) -> models.AssertionStdOperatorClass:
        condition_enum = (
            condition
            if isinstance(condition, SqlAssertionCondition)
            else SqlAssertionCondition(condition)
        )

        condition_to_operator = {
            SqlAssertionCondition.IS_EQUAL_TO: models.AssertionStdOperatorClass.EQUAL_TO,
            SqlAssertionCondition.IS_NOT_EQUAL_TO: models.AssertionStdOperatorClass.NOT_EQUAL_TO,
            SqlAssertionCondition.IS_GREATER_THAN: models.AssertionStdOperatorClass.GREATER_THAN,
            SqlAssertionCondition.IS_LESS_THAN: models.AssertionStdOperatorClass.LESS_THAN,
            SqlAssertionCondition.IS_WITHIN_A_RANGE: models.AssertionStdOperatorClass.BETWEEN,
            SqlAssertionCondition.GROWS_AT_MOST_ABSOLUTE: models.AssertionStdOperatorClass.LESS_THAN_OR_EQUAL_TO,
            SqlAssertionCondition.GROWS_AT_MOST_PERCENTAGE: models.AssertionStdOperatorClass.LESS_THAN_OR_EQUAL_TO,
            SqlAssertionCondition.GROWS_AT_LEAST_ABSOLUTE: models.AssertionStdOperatorClass.GREATER_THAN_OR_EQUAL_TO,
            SqlAssertionCondition.GROWS_AT_LEAST_PERCENTAGE: models.AssertionStdOperatorClass.GREATER_THAN_OR_EQUAL_TO,
            SqlAssertionCondition.GROWS_WITHIN_A_RANGE_ABSOLUTE: models.AssertionStdOperatorClass.BETWEEN,
            SqlAssertionCondition.GROWS_WITHIN_A_RANGE_PERCENTAGE: models.AssertionStdOperatorClass.BETWEEN,
        }

        if condition_enum not in condition_to_operator:
            raise SDKUsageError(f"Unknown condition: {condition_enum}")

        return condition_to_operator[condition_enum]  # type: ignore[return-value]

    @classmethod
    def get_change_type_from_condition(
        cls, condition: Union[SqlAssertionCondition, str]
    ) -> Optional[models.AssertionValueChangeTypeClass]:
        condition_enum = (
            condition
            if isinstance(condition, SqlAssertionCondition)
            else SqlAssertionCondition(condition)
        )

        if condition_enum in [
            SqlAssertionCondition.GROWS_AT_MOST_ABSOLUTE,
            SqlAssertionCondition.GROWS_AT_LEAST_ABSOLUTE,
            SqlAssertionCondition.GROWS_WITHIN_A_RANGE_ABSOLUTE,
        ]:
            return models.AssertionValueChangeTypeClass.ABSOLUTE  # type: ignore[return-value]
        elif condition_enum in [
            SqlAssertionCondition.GROWS_AT_MOST_PERCENTAGE,
            SqlAssertionCondition.GROWS_AT_LEAST_PERCENTAGE,
            SqlAssertionCondition.GROWS_WITHIN_A_RANGE_PERCENTAGE,
        ]:
            return models.AssertionValueChangeTypeClass.PERCENTAGE  # type: ignore[return-value]
        else:
            # Value-based conditions don't have a change type
            return None

    @classmethod
    def get_parameters(
        cls,
        parameters: Union[
            Union[float, int], tuple[Union[float, int], Union[float, int]]
        ],
    ) -> models.AssertionStdParametersClass:
        if isinstance(parameters, (float, int)):
            return models.AssertionStdParametersClass(
                value=models.AssertionStdParameterClass(
                    type=models.AssertionStdParameterTypeClass.NUMBER,
                    value=str(parameters),
                ),
            )
        elif isinstance(parameters, tuple):
            return models.AssertionStdParametersClass(
                minValue=models.AssertionStdParameterClass(
                    type=models.AssertionStdParameterTypeClass.NUMBER,
                    value=str(min(parameters[0], parameters[1])),
                ),
                maxValue=models.AssertionStdParameterClass(
                    type=models.AssertionStdParameterTypeClass.NUMBER,
                    value=str(max(parameters[0], parameters[1])),
                ),
            )
        else:
            raise SDKUsageError(
                "Either a single value or a range must be provided for the parameters of SqlAssertionCriteria"
            )

    @staticmethod
    def validate(criteria: "SqlAssertionCriteria") -> "SqlAssertionCriteria":
        operator = SqlAssertionCriteria.get_operator_from_condition(criteria.condition)
        condition_enum = (
            criteria.condition
            if isinstance(criteria.condition, SqlAssertionCondition)
            else SqlAssertionCondition(criteria.condition)
        )

        if operator in SINGLE_VALUE_NUMERIC_OPERATORS:
            if not isinstance(criteria.parameters, float) and not isinstance(
                criteria.parameters, int
            ):
                raise SDKUsageError(
                    f"The parameter value of SqlAssertionCriteria must be a single value numeric for condition {str(condition_enum)}"
                )
        elif operator in RANGE_OPERATORS:
            if not isinstance(criteria.parameters, tuple):
                raise SDKUsageError(
                    f"The parameter value of SqlAssertionCriteria must be a tuple range for condition {str(condition_enum)}"
                )
        else:
            raise SDKUsageError(
                f"Condition {str(condition_enum)} is not supported for SqlAssertionCriteria"
            )

        return criteria


class _SqlAssertionInput(_AssertionInput):
    def _assertion_type(self) -> str:
        """Get the assertion type."""
        return models.AssertionTypeClass.SQL

    def __init__(
        self,
        *,
        # Required fields
        dataset_urn: Union[str, DatasetUrn],
        entity_client: EntityClient,  # Needed to get the schema field spec for the detection mechanism if needed
        urn: Optional[Union[str, AssertionUrn]] = None,
        criteria: SqlAssertionCriteria,
        statement: str,
        # Optional fields
        display_name: Optional[str] = None,
        enabled: bool = True,
        schedule: Optional[Union[str, models.CronScheduleClass]] = None,
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes] = None,
        tags: Optional[TagsInputType] = None,
        created_by: Union[str, CorpUserUrn],
        created_at: datetime,
        updated_by: Union[str, CorpUserUrn],
        updated_at: datetime,
    ):
        _AssertionInput.__init__(
            self,
            dataset_urn=dataset_urn,
            entity_client=entity_client,
            urn=urn,
            display_name=display_name,
            enabled=enabled,
            schedule=schedule,
            incident_behavior=incident_behavior,
            tags=tags,
            source_type=models.AssertionSourceTypeClass.NATIVE,  # Native assertions are of type native, not inferred
            created_by=created_by,
            created_at=created_at,
            updated_by=updated_by,
            updated_at=updated_at,
        )
        self.criteria = criteria
        self.statement = statement

    def _create_monitor_info(
        self,
        assertion_urn: AssertionUrn,
        status: models.MonitorStatusClass,
        schedule: models.CronScheduleClass,
    ) -> models.MonitorInfoClass:
        """
        Create a MonitorInfoClass with all the necessary components.
        """
        return models.MonitorInfoClass(
            type=models.MonitorTypeClass.ASSERTION,
            status=status,
            assertionMonitor=models.AssertionMonitorClass(
                assertions=[
                    models.AssertionEvaluationSpecClass(
                        assertion=str(assertion_urn),
                        schedule=schedule,
                        parameters=self._get_assertion_evaluation_parameters(
                            "None",
                            None,  # Not used for sql assertions
                        ),
                    )
                ]
            ),
        )

    def _create_assertion_info(
        self,
        filter: Optional[models.DatasetFilterClass],  # Not used for sql assertions
    ) -> AssertionInfoInputType:
        """
        Create a SqlAssertionInfoClass for a sql assertion.

        Returns:
            A SqlAssertionInfoClass configured for sql.
        """
        SqlAssertionCriteria.validate(self.criteria)

        return models.SqlAssertionInfoClass(
            type=SqlAssertionCriteria.get_type_from_condition(self.criteria.condition),
            entity=str(self.dataset_urn),
            statement=self.statement,
            changeType=SqlAssertionCriteria.get_change_type_from_condition(
                self.criteria.condition
            ),
            operator=SqlAssertionCriteria.get_operator_from_condition(
                self.criteria.condition
            ),
            parameters=SqlAssertionCriteria.get_parameters(self.criteria.parameters),
        )

    def _get_assertion_evaluation_parameters(
        self,
        source_type: str,  # Not used for sql assertions
        field: Optional[FieldSpecType],  # Not used for sql assertions
    ) -> models.AssertionEvaluationParametersClass:
        return models.AssertionEvaluationParametersClass(
            type=models.AssertionEvaluationParametersTypeClass.DATASET_SQL,
        )

    # Not used for sql assertions
    def _convert_assertion_source_type_and_field(
        self,
    ) -> tuple[str, Optional[FieldSpecType]]:
        return "None", None

    # Not used for sql assertions
    def _create_filter_from_detection_mechanism(
        self,
    ) -> Optional[models.DatasetFilterClass]:
        return None

    def _convert_schedule(self) -> models.CronScheduleClass:
        """Create a schedule for a sql assertion.

        Returns:
            A CronScheduleClass with appropriate schedule settings.
        """
        if self.schedule is None:
            return DEFAULT_EVERY_SIX_HOURS_SCHEDULE

        return models.CronScheduleClass(
            cron=self.schedule.cron,
            timezone=self.schedule.timezone,
        )
